/*****************************************************
WinAvr
Chip type               : ATMega8
AVR Core Clock frequency: 4 MHz
Project: countdown timer
*****************************************************/
#include <avr/io.h>
#include <avr/power.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>

typedef unsigned char u8;
typedef unsigned int u16;
typedef unsigned long u32;


#define cpu_freq 4000000

typedef struct Bits_t
{
u8 b0 :1;
u8 b1 :1;
u8 b2 :1;
u8 b3 :1;
u8 b4 :1;
u8 b5 :1;
u8 b6 :1;
u8 b7 :1;
}Bits;

#define prtb (*((volatile Bits*)&PORTB))
#define pinb (*((volatile Bits*)&PINB))
#define prtc (*((volatile Bits*)&PORTC))
#define pinc (*((volatile Bits*)&PINC))
#define prtd (*((volatile Bits*)&PORTD))
#define pind (*((volatile Bits*)&PIND))

#define key0 pinb.b0
#define key1 pinb.b1
#define key2 pinb.b2
#define key3 pinb.b3
#define key4 pinb.b4
#define key5 pinb.b5

#define led0 prtc.b0
#define led1 prtc.b1
#define led2 prtc.b2
#define led3 prtc.b3
#define led4 prtc.b4
#define led5 prtc.b5

#define relay prtd.b7

enum {rls, prs, hold} btn0, btn1, btn2, btn3, btn4, btn5;
enum {run, work, wait, sdn} dev_state;

static u8 fg[6]={0,0,0,0,0,0};
u16 tim[3];

static u32 time=0;
static u8 t_run=0;

void gpio_init(void)
{
 PORTB=PORTC=PORTD=0; DDRC=DDRD=0xff;
 DDRB=0; PORTB=0xff; 
}


void t2_init(void)
{
 TCCR2 = 1<<WGM21|2; //CTC, OCR2 top, /8
 OCR2  = (u8)(cpu_freq/8/2000)-1;
 TIMSK|= 1<<TOIE2;
}
void scan_keys(void)
{
#define lp  4
#define lh  10
#define lr  2

static u8 count[12];
static u8 tkey[6];

for(u8 i=6;i;i--) if(tkey[i-1])tkey[i-1]--;

if ((!key0)&&(!tkey[0])) {count[0]++;tkey[0]=lr;}
  else if((key0)&&(!tkey[0])) {count[0]=count[1]=0;btn0=rls;}
if ((count[0]>lp)&&(btn0!=hold)) {btn0=prs; count[0]=0; count[1]++;}
//if (count[1]>lh) {btn0=hold; count[1]=0;}

if ((!key1)&&(!tkey[1])) {count[2]++;tkey[1]=lr;}
  else if((key1)&&(!tkey[1])) {count[3]=count[2]=0;btn1=rls;}
if ((count[2]>lp)&&(btn1!=hold)) {btn1=prs; count[2]=0; count[3]++;}
//if (count[3]>lh) {btn1=hold; count[3]=0;}

if ((!key2)&&(!tkey[2])) {count[4]++;tkey[2]=lr;}
  else if((key2)&&(!tkey[2])) {count[5]=count[4]=0;btn2=rls;}
if ((count[4]>0xf0)&&(btn2!=hold)) {btn2=prs; count[4]=0; count[5]++;}
//if (count[5]>lh) {btn2=hold; count[5]=0;}

if ((!key3)&&(!tkey[3])) {count[6]++;tkey[3]=lr;}
  else if((key3)&&(!tkey[3])) {count[7]=count[6]=0;btn3=rls;}
if ((count[6]>lp)&&(btn3!=hold)) {btn3=prs; count[6]=0; count[7]++;}
//if (count[1]>lh) {btn0=hold; count[1]=0;}

if ((!key4)&&(!tkey[4])) {count[8]++;tkey[4]=lr;}
  else if((key4)&&(!tkey[4])) {count[9]=count[8]=0;btn4=rls;}
if ((count[8]>lp)&&(btn4!=hold)) {btn4=prs; count[8]=0; count[9]++;}
//if (count[3]>lh) {btn1=hold; count[3]=0;}

if ((!key5)&&(!tkey[5])) {count[10]++;tkey[5]=lr;}
  else if((key5)&&(!tkey[5])) {count[11]=count[10]=0;btn5=rls;}
if ((count[10]>0xf0)&&(btn5!=hold)) {btn5=prs; count[10]=0; count[11]++;}
//if (count[5]>lh) {btn2=hold; count[5]=0;}

}



ISR (TIMER2_OVF_vect)
{
	if(t_run){if(time){time--;relay=1;}else relay=0;}
	else relay = 0;

 
  for(u8 i=0;i<3;i++)if(tim[i])tim[i]--;
	wdt_reset();
  prtd.b0^=1;
}


void devcon(void), keywork(void);

int main (void)
{
 gpio_init();
 //t0_init();
 //t1_init();
 t2_init();

sei();
wdt_enable(WDTO_500MS);

w___:
{

 keywork();
 devcon();
 scan_keys();

}
goto w___;
return 1;
}

void keywork(void)
{
 static u8 trg[6]={0,0,0,0,0,0};
 
	if(btn0==prs&&!trg[0]){trg[0]=1;fg[0]^=1;}
	else if(btn0==rls&&trg[0]){trg[0]=0;if(fg[0])led0=1;else led0=0;}
	if(btn1==prs&&!trg[1]){trg[1]=1;fg[1]^=1;}
	else if(btn1==rls&&trg[1]){trg[1]=0;if(fg[1])led1=1;else led1=0;}
	if(btn2==prs&&!trg[2]){trg[2]=1;fg[2]^=1;}
	else if(btn2==rls&&trg[2]){trg[2]=0;if(fg[2])led2=1;else led2=0;}  
	if(btn3==prs&&!trg[3]){trg[3]=1;fg[3]^=1;}
	else if(btn3==rls&&trg[3]){trg[3]=0;if(fg[3])led3=1;else led3=0;}
	if(btn4==prs&&!trg[4]){trg[4]=1;fg[4]^=1;}
	else if(btn4==rls&&trg[4]){trg[4]=0;if(fg[4])led4=1;else led4=0;}
	if(btn5==prs&&!trg[5]){trg[5]=1;fg[5]^=1;}
	else if(btn5==rls&&trg[5]){trg[5]=0;if(fg[5])led5=1;else led5=0;} 
}

void devcon(void)
{
 static u8 itr, ttr[6]={0,0,0,0,0,0}, time_tmp;
 
 switch(dev_state)
 {
	
	
	case run:
	{
	 if(!itr)
	 {
		itr=1; time_tmp=0; time=0; t_run=0; 
		tim[0]=1000;
		PORTC = 0xff;
		while(tim[0]);
		PORTC = 0;
		fg[5]=fg[4]=fg[3]=fg[2]=fg[1]=fg[0]=0;
		ttr[5]=ttr[4]=ttr[3]=ttr[2]=ttr[1]=ttr[0]=0;
	 }
	 
	 if(fg[0])dev_state=work;
	 if(fg[1]&&!ttr[1]){ttr[1]=1; time_tmp+=3;}
	 else if(!fg[1]&&ttr[1]){ttr[1]=0;if(time_tmp>=3)time_tmp-=3;else time_tmp=0;}
	 if(fg[2]&&!ttr[2]){ttr[2]=1; time_tmp+=6;}
	 else if(!fg[2]&&ttr[2]){ttr[2]=0;if(time_tmp>=6)time_tmp-=6;else time_tmp=0;}
	 if(fg[3]&&!ttr[3]){ttr[3]=1; time_tmp+=12;}
	 else if(!fg[3]&&ttr[3]){ttr[3]=0;if(time_tmp>=12)time_tmp-=12;else time_tmp=0;}
	 if(fg[4]&&!ttr[4]){ttr[4]=1; time_tmp+=18;}
	 else if(!fg[4]&&ttr[4]){ttr[4]=0;if(time_tmp>=18)time_tmp-=18;else time_tmp=0;}
	 if(fg[5]&&!ttr[5]){ttr[5]=1; time_tmp+=24;}
	 else if(!fg[5]&&ttr[5]){ttr[5]=0;if(time_tmp>=24)time_tmp-=24;else time_tmp=0;}
	
	if(dev_state!=run) itr=0;
	}	
	break;

	case work:
	{
	 static u8 state_tmp;
	 if(!itr)
	 {
		itr=1; 
		cli();
		time=(u32)time_tmp*10*60*2000;
		//time=(u32)time_tmp*2000;
		sei();
		state_tmp = fg[5]<<5|fg[4]<<4|fg[3]<<3|fg[2]<<2|fg[1]<<1|fg[0];
		t_run=1;
	 }
	 	
		if(!tim[1])
		{
		 tim[1]=200;
		 if(!time)dev_state=run;
		 if(state_tmp!=(fg[5]<<5|fg[4]<<4|fg[3]<<3|fg[2]<<2|fg[1]<<1|fg[0]))
		 {
			t_run=0; time=0;
			state_tmp = fg[5]<<5|fg[4]<<4|fg[3]<<3|fg[2]<<2|fg[1]<<1|fg[0];
			
			if(!fg[0])dev_state=run;
			if(fg[1]&&!ttr[1]){ttr[1]=1; time_tmp+=3;}
				else if(!fg[1]&&ttr[1]){ttr[1]=0;if(time_tmp>=3)time_tmp-=3;else time_tmp=0;}
			if(fg[2]&&!ttr[2]){ttr[2]=1; time_tmp+=6;}
				else if(!fg[2]&&ttr[2]){ttr[2]=0;if(time_tmp>=6)time_tmp-=6;else time_tmp=0;}
			if(fg[3]&&!ttr[3]){ttr[3]=1; time_tmp+=12;}
				else if(!fg[3]&&ttr[3]){ttr[3]=0;if(time_tmp>=12)time_tmp-=12;else time_tmp=0;}
			if(fg[4]&&!ttr[4]){ttr[4]=1; time_tmp+=18;}
				else if(!fg[4]&&ttr[4]){ttr[4]=0;if(time_tmp>=18)time_tmp-=18;else time_tmp=0;}
			if(fg[5]&&!ttr[5]){ttr[5]=1; time_tmp+=24;}
				else if(!fg[5]&&ttr[5]){ttr[5]=0;if(time_tmp>=24)time_tmp-=24;else time_tmp=0;}
		
				cli();
				time=(u32)time_tmp*10*60*2000;
				//time=(u32)time_tmp*2000;
				sei();
				t_run=1;
		 }
		}
	
	if(dev_state!=work) itr=0;
	}	
	break;

	case wait:
	{
	
	
	if(dev_state!=wait) itr=0;
	}	
	break;	
	
	case sdn:
	{
	
	
	if(dev_state!=sdn) itr=0;
	}	
	break;	
 
 }



}